/* $Id: ECalcView.java,v 1.11 2006/03/25 14:09:59 gsson Exp $ */

/*
 * Copyright (c) 2006 Henrik Gustafsson <henrik.gustafsson@fnord.se>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package se.fnord.ecalc.views;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringReader;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Properties;

import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.forms.widgets.FormText;
import org.eclipse.ui.forms.widgets.FormToolkit;
import org.eclipse.ui.forms.widgets.ScrolledForm;
import org.eclipse.ui.forms.widgets.ScrolledFormText;
import org.eclipse.ui.part.ViewPart;

import se.fnord.ecalc.ECalcPlugin;
import se.fnord.ecalc.parser.EvaluationVisitor;
import se.fnord.ecalc.parser.ExpressionContext;
import se.fnord.ecalc.parser.ExpressionException;
import se.fnord.ecalc.parser.ExpressionParser;
import se.fnord.ecalc.parser.ExpressionParserVisitor;
import se.fnord.ecalc.parser.IValue;
import se.fnord.ecalc.parser.Literal;
import se.fnord.ecalc.parser.ParseException;
import se.fnord.ecalc.parser.SimpleNode;
import se.fnord.ecalc.parser.TokenMgrError;
import se.fnord.ecalc.parser.functions.Ceiling;
import se.fnord.ecalc.parser.functions.Cosine;
import se.fnord.ecalc.parser.functions.Floor;
import se.fnord.ecalc.parser.functions.Hypotenuse;
import se.fnord.ecalc.parser.functions.Random;
import se.fnord.ecalc.parser.functions.Round;
import se.fnord.ecalc.parser.functions.Sine;
import se.fnord.ecalc.parser.functions.Tangent;

public class ECalcView extends ViewPart implements SelectionListener,
		KeyListener, ModifyListener {

	private FormToolkit toolkit;
	private ScrolledForm form;
	private Text input;
	private FormText resultsView;
	private ScrolledFormText resultsContainer;

	private List calculations = new LinkedList();
	private List history = new LinkedList();
	private int history_position = -1;

	private ExpressionContext context = new ExpressionContext();

	private Properties messages = new Properties();

	class Calculation {
		private String expression;

		private String result;

		public Calculation(String expression, String result) {
			this.expression = expression;
			this.result = result;
		}

		public String getExpression() {
			return this.expression;
		}

		public String getResult() {
			return this.result;
		}
	}

	public ECalcView() {
		context.setFunction("sin", new Sine());
		context.setFunction("cos", new Cosine());
		context.setFunction("tan", new Tangent());
		context.setFunction("round", new Round());
		context.setFunction("ceil", new Ceiling());
		context.setFunction("floor", new Floor());
		context.setFunction("random", new Random());
		context.setFunction("hypot", new Hypotenuse());
		loadMessages();
		restoreState();
	}

	public void createPartControl(Composite parent) {
		toolkit = new FormToolkit(parent.getDisplay());
		form = toolkit.createScrolledForm(parent);
		GridLayout l = new GridLayout(1, true);
		form.getBody().setLayout(l);

		resultsContainer = new ScrolledFormText(form.getBody(), false);

		resultsView = toolkit.createFormText(resultsContainer, false);
		resultsView.setParagraphsSeparated(true);

		resultsContainer.setMinWidth(100);
		resultsContainer.setExpandHorizontal(true);
		resultsContainer.setMinHeight(100);
		resultsContainer.setExpandVertical(true);

		resultsContainer.setAlwaysShowScrollBars(false);
		resultsContainer.setFormText(resultsView);
		toolkit.adapt(resultsContainer);
		resultsContainer.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true,
				true));

		Label separator = toolkit.createSeparator(form.getBody(),
				SWT.HORIZONTAL);
		separator
				.setLayoutData(new GridData(SWT.FILL, SWT.BOTTOM, true, false));

		input = toolkit.createText(form.getBody(), "");
		input.setLayoutData(new GridData(SWT.FILL, SWT.BOTTOM, true, false));
		input.addSelectionListener(this);
		input.addKeyListener(this);
		input.addModifyListener(this);
		form.reflow(true);

		buildResults();
	}

	private void loadMessages() {
		ECalcPlugin plugin = ECalcPlugin.getDefault();
		try {
			messages.load(plugin.openStream(new Path("messages.properties"),
					true));

		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void restoreState() {
		ECalcPlugin plugin = ECalcPlugin.getDefault();
		IPath state_location = plugin.getStateLocation();

		// TODO: Restore arithmetic-type

		history.clear();
		try {
			File histfile = state_location.addTrailingSeparator().append(
					"history.txt").toFile();
			BufferedReader br = new BufferedReader(new FileReader(histfile));
			String s = br.readLine();
			while (s != null) {
				history.add(s);
				s = br.readLine();
			}
			br.close();
			history_position = -1;
		} catch (IOException e) {
			// TODO: Don't fail silently if the error is bad :)
		}

		calculations.clear();
		try {
			File calcfile = state_location.addTrailingSeparator().append(
					"calculations.txt").toFile();
			BufferedReader br = new BufferedReader(new FileReader(calcfile));
			String expr = br.readLine();
			String result = br.readLine();
			while (expr != null && result != null) {
				calculations.add(new Calculation(expr, result));
				expr = br.readLine();
				result = br.readLine();
			}
			br.close();
		} catch (IOException e) {
			// TODO: Don't fail silently if the error is bad :)
		}
	}

	private void storeState() {
		ECalcPlugin plugin = ECalcPlugin.getDefault();
		IPath state_location = plugin.getStateLocation();

		// TODO: Store arithmetic-type

		try {
			File histfile = state_location.addTrailingSeparator().append(
					"history.txt").toFile();

			BufferedWriter bw = new BufferedWriter(new FileWriter(histfile));

			ListIterator i = history.listIterator();
			while (i.hasNext()) {
				String c = (String) i.next();
				bw.write(c);
				bw.newLine();
			}
			bw.close();
		} catch (IOException e) {
			// TODO: Don't fail silently if the error is bad :)
		}

		try {
			File calcfile = state_location.addTrailingSeparator().append(
					"calculations.txt").toFile();
			BufferedWriter bw = new BufferedWriter(new FileWriter(calcfile));

			ListIterator i = calculations.listIterator();
			while (i.hasNext()) {
				Calculation c = (Calculation) i.next();
				bw.write(c.getExpression());
				bw.newLine();
				bw.write(c.getResult());
				bw.newLine();
			}
			bw.close();
		} catch (IOException e) {
			// TODO: Don't fail silently if the error is bad :)
		}
	}

	public void dispose() {
		storeState();
		super.dispose();
	}

	public void setFocus() {
		input.setFocus();
	}

	private void buildResults() {
		String s = "<form>";
		ListIterator i = calculations.listIterator();
		while (i.hasNext()) {
			Calculation c = (Calculation) i.next();
			String result = c.getResult();
			String expression = c.getExpression();
			if (result.startsWith("%")) {
				result = messages.getProperty(result.substring(1), result);
			}

			result = result.replaceAll("&", "&amp;").replaceAll("<", "&lt;")
					.replaceAll(">", "&gt;");
			expression = expression.replaceAll("&", "&amp;").replaceAll("<",
					"&lt;").replaceAll(">", "&gt;");
			s += "<p vspace='true'>" + expression + "</p>";
			s += "<li vspace='false' style='text' value='&gt;' bindent='0' indent='50'><b>"
					+ result + "</b></li>";
		}

		s += "</form>";
		resultsView.setText(s, true, false);
		resultsContainer.reflow(true);
		resultsContainer.setOrigin(0, resultsView.getBounds().height);

	}

	public void widgetSelected(SelectionEvent e) {}

	public void widgetDefaultSelected(SelectionEvent e) {

		ExpressionParser parser = null;
		String expression = input.getText();
		String result;
		if (expression.matches(".*\\S+.*")) {
			try {
				parser = new ExpressionParser(new StringReader(expression));
				SimpleNode parseTree = parser.Start();
				ExpressionParserVisitor visitor = new EvaluationVisitor();
				IValue value = parseTree.accept(visitor, context);
				Literal l = new Literal();
				l.fromIValue(value);
				context.setVariable("ans", l);
				result = "" + value;
			} catch (ExpressionException exception) {
				result = exception.getMessageId();
			} catch (ParseException exception) {
				result = "%se.fnord.ecalc.exceptions.parseexception";
			} catch (TokenMgrError exception) {
				result = "%se.fnord.ecalc.exceptions.parseexception";
			} catch (Throwable exception) {
				result = "Some kind of error-thingy";
			}

			addCalculation(expression, result);
			buildResults();
			history_position = history.size() - 1;
			if (history_position != -1) {
				setFromHistory();
				input.setSelection(0, expression.length());
			}
		}
	}

	public void setForceArith(int type) {
		context.arithmeticType = type;
	}

	private void addCalculation(String expression, String result) {
		if (calculations.size() == 50) {
			calculations.remove(0);
		}
		calculations.add(new Calculation(expression, result));
		addHistory(expression);
	}

	private boolean addHistory(String expression) {
		if (expression.matches(".*\\S+.*")) {
			ListIterator i = history.listIterator();
			while (i.hasNext()) {
				String c = (String) i.next();
				if (c.equals(expression)) {
					i.remove();
					break;
				}
			}

			if (history.size() == 50) {
				history.remove(0);
			}
			history.add(expression);
			return true;
		}
		else {
			return false;
		}
	}

	public void clearHistory() {
		calculations.clear();
		buildResults();
	}

	private void setFromHistory() {
		String entry = (String) history.get(history_position);
		input.setText(entry);
		input.setSelection(entry.length() + 1);
	}

	public void keyPressed(KeyEvent e) {
		if (e.keyCode == SWT.ARROW_UP) {
			e.doit = false;

			if (history_position == -1) {
				addHistory(input.getText());
				history_position = history.size() - 1;
				setFromHistory();
			}
			else if (history.size() > 0 && history_position > 0) {
				history_position--;
				setFromHistory();
			}
		}
		else if (e.keyCode == SWT.ARROW_DOWN) {
			e.doit = false;
			if (history_position == -1) {
				addHistory(input.getText());
				input.setText("");
			}
			else if (history_position < history.size() - 1) {
				history_position++;
				setFromHistory();
			}
			else {
				addHistory(input.getText());
				input.setText("");
				history_position = -1;
			}
		}
	}

	public void keyReleased(KeyEvent e) {}

	public void modifyText(ModifyEvent e) {
		// TODO: Really ugly construct. There must be a more elegant way of
		// doing this.
		if (history_position != -1
				&& !input.getText().equals(history.get(history_position)))
			history_position = -1;
	}

}